/*
Module : AppSoundsManager.cpp
Purpose: Implementation for the CAppSoundsManager class
Created: PJN / 22-09-2006
History: PJN / 23-09-2006 1. The existing APP_SOUND macro has been renamed to APP_SOUND_FROM_RESOURCE.
                          2. Now supports registration of file based wave files using the APP_SOUND_FROM_FILE macro.
                          3. Now supports the concept of a disabled sound, where the sound will only be played if the 
                          end-user configures the value in the control panel through the new APP_SOUND_DEFAULT_NONE macro.
         PJN / 14-06-2007 1. Updated Register and UnRegister methods to preserve last error code if it cannot open the 
                          "HKCU\AppEvents\EventLabels" registry key.
                          2. Updated copyright details.
         PJN / 04-01-2009 1. Updated copyright details.
                          2. Removed VC 6 style AppWizard comments from the code
                          3. Removed unused CAppSoundsManager destructor
                          4. The code has now been updated to support VC 2005 or later only
                          5. Fixed a potential buffer overwrite in CAppSoundsManager::GetCurrentSoundFromRegistry. This
                          bug was spotted with Code Analysis.
                          6. Code now compiles cleanly using Code Analysis (/analyze)
                          7. The Register method has been completely rewritten and refactored
                          8. The Unregister method now uses SHDeleteKey to ensure key values which have subkeys are 
                          cleanly removed from the registry. The code in this method has also been refactored.
         PJN / 24-01-2016 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2008 - 2015
                          3. The IsRegistered method now use ATL:CRegKey and returns a HRESULT return value
                          4. The GetCurrentSoundFromRegistry method now uses ATL::CRegKey and returns a 
                          HRESULT return value
                          5. The Register method now uses ATL::CRegKey and returns a HRESULT return value
                          6. The RegisterSound method now uses ATL::CRegKey and returns a HRESULT return 
                          value
                          7. The UnRegister method now uses ATL::CRegKey and returns a HRESULT return value
                          8. The Initialize method now returns a HRESULT return value
                          9. The CAppSoundsManager::UnRegister method now uses ATL::CRegKey::RecurseDeleteKey instead of SHDeleteKey.
                          10. Added SAL annotations to all the code.
                          11. Reworked the classes to optionally compile without MFC. By default the class now use STL classes and idioms 
                          but if you define CAPPSOUNDSMANAGER_MFC_EXTENSIONS the class will revert back to the MFC behaviour.
         PJN / 02-12-2017 1. Updated copyright details.
                          2. Replaced NULL throughout the codebase with nullptr. This means that the minimum requirement 
                          for the framework is now VC 2010.
                          3. Replaced BOOL throughout the codebase with bool.
                          4. Replaced CString::operator LPC*STR() calls throughout the codebase with CString::GetString calls
         PJN / 20-10-2018 1. Updated copyright details
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that the code will
                          now only compile on VC 2017 or later.
                          3. Removed all code which supported CAPPSOUNDSMANAGER_MFC_EXTENSIONS code path
         PJN / 14-03-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 04-05-2020 1. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 24-04-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations.

Copyright (c) 2006 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"

#ifndef _INC_MMSYSTEM
#pragma message("To avoid this message, please put mmsystem.h in your pre compiled header (normally stdafx.h)")
#include <mmsystem.h>
#endif //#ifndef _INC_MMSYSTEM

#include "AppSoundsManager.h"


//////////////////// Macros / Defines /////////////////////////////////////////

#pragma comment(lib, "winmm.lib") //Automatically pull in support for the ::PlaySound function which we use


//////////////////// Implementation ///////////////////////////////////////////

CAppSoundsManager::CAppSoundsManager() noexcept : m_nSounds{0},
                                                  m_pSounds{nullptr},
                                                  m_bEnabled{true},
                                                  m_hResource{nullptr},
                                                  m_nAppFriendlyNameID{0}
{
}

HRESULT CAppSoundsManager::Initialize(_In_z_ LPCTSTR pszAppFriendlyName, _In_ UINT nAppFriendlyNameID, _In_ HINSTANCE hResource)
{
  m_sAppFriendlyName = pszAppFriendlyName;
  m_nAppFriendlyNameID = nAppFriendlyNameID;
  m_hResource = hResource;

  //Also SetAppFileName to be the filename for this exe
  ATL::CAtlString sApp;
  const DWORD dwGMFN{GetModuleFileName(nullptr, sApp.GetBuffer(_MAX_PATH), _MAX_PATH)};
  sApp.ReleaseBuffer();
  if (dwGMFN == 0)
    return ATL::AtlHresultFromLastError();
  ATL::CAtlString sFname;
  _tsplitpath_s(sApp, nullptr, 0, nullptr, 0, sFname.GetBuffer(_MAX_FNAME), _MAX_FNAME, nullptr, 0);
  sFname.ReleaseBuffer();
  SetAppFileName(sFname);

  //Finally call the "SetupSoundsTable" method which should be setup for us via END_APP_SOUNDS
  SetupSoundsTable();

  return S_OK;
}

void CAppSoundsManager::SetupSounds(_In_ const CAppSound* pSounds, _In_ int nSounds) noexcept
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(pSounds != nullptr);
#pragma warning(suppress: 26477)
  ATLASSERT(nSounds);

  //Hive away the values into member variables
  m_pSounds = pSounds;
  m_nSounds = nSounds;
}

ATL::CAtlString CAppSoundsManager::GetAppKeyName()
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_sFilename.GetLength()); //Did you forget to call Initialize or SetAppFileName

  //Form what will be the return value from this function
  ATL::CAtlString sAppKey(_T("AppEvents\\Schemes\\Apps\\"));
  sAppKey += m_sFilename;

  return sAppKey;
}

ATL::CAtlString CAppSoundsManager::GetSoundAliasCurrentValueKeyName(_In_ UINT nSoundID)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_sFilename.GetLength()); //Did you forget to call Initialize or SetAppFileName

  ATL::CAtlString sCurrentValue;
  sCurrentValue.Format(_T("%s\\%s_%u\\.current"), GetAppKeyName().GetString(), m_sFilename.GetString(), nSoundID);
  return sCurrentValue;
}

ATL::CAtlString CAppSoundsManager::FormAlias(_In_ const CAppSound& sound)
{
  ATL::CAtlString sAlias;
  sAlias.Format(_T("%s_%u"), m_sFilename.GetString(), sound.m_nID);
  return sAlias;
}

ATL::CAtlString CAppSoundsManager::FormIndirectShellString(_In_z_ LPCTSTR pszResourceModule, _In_ UINT nID)
{
  ATL::CAtlString sIndirectShellString;
  sIndirectShellString.Format(_T("@\"%s\",-%u"), pszResourceModule, nID);
  return sIndirectShellString;
}

const CAppSound* CAppSoundsManager::FindAppSound(_In_ UINT nSoundID) noexcept
{
  for (int i{0}; i<m_nSounds; i++)
  {
#pragma warning(suppress: 26477)
    ATLASSUME(m_pSounds != nullptr);
#pragma warning(suppress: 26481)
    if (m_pSounds[i].m_nID == nSoundID)
#pragma warning(suppress: 26481)
      return &(m_pSounds[i]);
  }

  return nullptr;
}

HRESULT CAppSoundsManager::GetCurrentSoundFromRegistry(_In_ UINT nSoundID, _Inout_ ATL::CAtlString& sSoundFileName)
{
  //Check to see if a sound file has been specified for the sound in the registry
  ATL::CRegKey aliasKey;
  LSTATUS lResult{aliasKey.Open(HKEY_CURRENT_USER, GetSoundAliasCurrentValueKeyName(nSoundID), KEY_READ)};
  if (lResult != ERROR_SUCCESS)
    return AtlHresultFromWin32(lResult);

  //Query the ".current" value
  DWORD dwBufferSize{0};
  DWORD dwType{0};
  lResult = aliasKey.QueryValue(nullptr, &dwType, nullptr, &dwBufferSize);
  if (lResult != ERROR_SUCCESS)
    return AtlHresultFromWin32(lResult);
  ATL::CAtlString sFilename;
  dwBufferSize = _MAX_PATH;
#pragma warning(suppress: 26446)
  lResult = aliasKey.QueryStringValue(nullptr, sFilename.GetBuffer(_MAX_PATH), &dwBufferSize);
  sFilename.ReleaseBuffer();
  if (lResult != ERROR_SUCCESS)
    return AtlHresultFromWin32(lResult);

  if (dwType == REG_SZ)
    sSoundFileName = sFilename;
  else if (dwType == REG_EXPAND_SZ)
  {
    const DWORD dwChars{ExpandEnvironmentStrings(sFilename, sSoundFileName.GetBuffer(_MAX_PATH), _MAX_PATH)};
    sSoundFileName.ReleaseBuffer();
    if (dwChars == 0)
      return AtlHresultFromLastError();
  }

  return S_OK;
}

HRESULT CAppSoundsManager::IsRegistered()
{
  //Check to see if all sounds have been registered under "AppEvents\Schemes\Apps\<FILENAME>"
  ATL::CRegKey appKey;
  LSTATUS lResult{appKey.Open(HKEY_CURRENT_USER, GetAppKeyName(), KEY_READ)};
  if (lResult != ERROR_SUCCESS)
    return ATL::AtlHresultFromWin32(lResult);

  //Check to see if all the sounds are registered
  for (int i{0}; i<m_nSounds; i++)
  {
#pragma warning(suppress: 26481)
    ATL::CAtlString sAlias{FormAlias(m_pSounds[i])};
    ATL::CRegKey soundKey;
    lResult = soundKey.Open(appKey, sAlias, KEY_READ);
    switch (lResult)
    {
      case ERROR_FILE_NOT_FOUND:
      {
        return S_FALSE;
        break;
      }
      case ERROR_SUCCESS:
      {
        break;
      }
      default:
      {
        return ATL::AtlHresultFromWin32(lResult);
        break;
      }
    }
  }

  return S_OK;
}

HRESULT CAppSoundsManager::RegisterSound(_In_ HKEY hAppKey, _In_ HKEY hEventLabelsKey, _In_ int nSoundIndex, _In_ bool bFlush, _In_z_ LPCTSTR pszAppPath, _In_z_ LPCTSTR pszResourceModule)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSUME(m_pSounds != nullptr);

  //Load up the friendly name for this app sound
  ATL::CAtlString sFriendlyName;
#pragma warning(suppress: 26481)
  if (!sFriendlyName.LoadString(m_hResource, m_pSounds[nSoundIndex].m_nID))
    return ATL::AtlHresultFromLastError();

  //Create the alias for this app sound
#pragma warning(suppress: 26481)
  ATL::CAtlString sAlias{FormAlias(m_pSounds[nSoundIndex])};

  //Create the values under "HKCU\AppEvents\Schemes\Apps\[Alias]"
  ATL::CRegKey soundKey;
#pragma warning(suppress: 26477)
  LSTATUS lResult{soundKey.Create(hAppKey, sAlias, REG_NONE, REG_OPTION_NON_VOLATILE, KEY_WRITE)};
  if (lResult != ERROR_SUCCESS)
    return ATL::AtlHresultFromWin32(lResult);
  lResult = soundKey.SetStringValue(nullptr, sFriendlyName, REG_SZ);
  if (lResult != ERROR_SUCCESS)
    return ATL::AtlHresultFromWin32(lResult);

  //Register the wave file path if required
#pragma warning(suppress: 26481)
  if (m_pSounds[nSoundIndex].m_pszPath && _tcslen(m_pSounds[nSoundIndex].m_pszPath))
  {
    ATL::CRegKey currentKey;
#pragma warning(suppress: 26477)
    lResult = currentKey.Create(soundKey, _T(".current"), REG_NONE, REG_OPTION_NON_VOLATILE, KEY_WRITE);
    if (lResult != ERROR_SUCCESS)
      return ATL::AtlHresultFromWin32(lResult);

#pragma warning(suppress: 26481)
    ATL::CAtlString sWaveFile{m_pSounds[nSoundIndex].m_pszPath};
    sWaveFile.Replace(_T("%APPPATH%"), pszAppPath);
    lResult = currentKey.SetStringValue(nullptr, sWaveFile, REG_SZ);
    if (lResult != ERROR_SUCCESS)
      return ATL::AtlHresultFromWin32(lResult);

    //Optionally flush
    if (bFlush)
    {
      lResult = currentKey.Flush();
      if (lResult != ERROR_SUCCESS)
        return ATL::AtlHresultFromWin32(lResult);
    }
  }

  //Create the values under "HKCU\AppEvents\EventLabels\[Alias]"
  ATL::CRegKey eventLabelKey;
#pragma warning(suppress: 26477)
  lResult = eventLabelKey.Create(hEventLabelsKey, sAlias, REG_NONE, REG_OPTION_NON_VOLATILE, KEY_WRITE);
  if (lResult != ERROR_SUCCESS)
    return ATL::AtlHresultFromWin32(lResult);

  lResult = eventLabelKey.SetStringValue(nullptr, sFriendlyName, REG_SZ);
  if (lResult != ERROR_SUCCESS)
    return ATL::AtlHresultFromWin32(lResult);

  //Also set the DispFileName value
#pragma warning(suppress: 26481)
  ATL::CAtlString sIndirectShellString{FormIndirectShellString(pszResourceModule, m_pSounds[nSoundIndex].m_nID)};
  lResult = eventLabelKey.SetStringValue(_T("DispFileName"), sIndirectShellString, REG_SZ);
  if (lResult != ERROR_SUCCESS)
    return ATL::AtlHresultFromWin32(lResult);

  if (bFlush)
  {
    lResult = eventLabelKey.Flush();
    if (lResult != ERROR_SUCCESS)
      return ATL::AtlHresultFromWin32(lResult);
    lResult = soundKey.Flush();
    if (lResult != ERROR_SUCCESS)
      return ATL::AtlHresultFromWin32(lResult);
  }

  return S_OK;
}

HRESULT CAppSoundsManager::Register(_In_ bool bFlush)
{
  //Get the full path to the module which contains our sound resources
  ATL::CAtlString sResourceModule;
#pragma warning(suppress: 26477)
  ATLASSERT(m_hResource); //Did you forget to call Initialize?
  DWORD dwGMFN{GetModuleFileName(m_hResource, sResourceModule.GetBuffer(_MAX_PATH), _MAX_PATH)};
  sResourceModule.ReleaseBuffer();
  if (dwGMFN == 0)
    return ATL::AtlHresultFromLastError();

  //Form the path for %APPPATH%
  ATL::CAtlString sAppPath;
#pragma warning(suppress: 26446)
  dwGMFN = GetModuleFileName(nullptr, sAppPath.GetBuffer(_MAX_PATH), _MAX_PATH);
  sAppPath.ReleaseBuffer();
  if (dwGMFN == 0)
    return ATL::AtlHresultFromLastError();
  ATL::CAtlString sDrive;
  ATL::CAtlString sDir;
  _tsplitpath_s(sAppPath, sDrive.GetBuffer(_MAX_DRIVE), _MAX_DRIVE, sDir.GetBuffer(_MAX_DIR), _MAX_DIR, nullptr, 0, nullptr, 0);
  sDir.ReleaseBuffer();
  sDrive.ReleaseBuffer();
  _tmakepath_s(sAppPath.GetBuffer(_MAX_PATH), _MAX_PATH, sDrive, sDir, nullptr, nullptr);
  sAppPath.ReleaseBuffer();

  const int nAppPathLength{sAppPath.GetLength()};
  if (nAppPathLength && (sAppPath[nAppPathLength - 1] == _T('\\')))
    sAppPath = sAppPath.Left(nAppPathLength - 1);

  //Try to create/open the HKCU\AppEvents\Schemes\Apps\<FILENAME> key
  ATL::CRegKey appKey;
#pragma warning(suppress: 26477)
  LSTATUS lResult{appKey.Create(HKEY_CURRENT_USER, GetAppKeyName(), REG_NONE, REG_OPTION_NON_VOLATILE, KEY_WRITE)};
  if (lResult != ERROR_SUCCESS)
    return ATL::AtlHresultFromWin32(lResult);

  //Also try to open the HKCU\AppEvents\EventLabels key
  ATL::CRegKey eventLabelsKey;
  lResult = eventLabelsKey.Open(HKEY_CURRENT_USER, _T("AppEvents\\EventLabels"), KEY_WRITE);
  if (lResult != ERROR_SUCCESS)
    return ATL::AtlHresultFromWin32(lResult);

  //Set the default value for the key opened above, This corresponds to the Application name which is shown in the Sounds control panel applet
  lResult = appKey.SetStringValue(nullptr, m_sAppFriendlyName, REG_SZ);
  if (lResult != ERROR_SUCCESS)
    return ATL::AtlHresultFromWin32(lResult);

  //Also add the DispFileName value also if required. This allows CAppSoundsManager to support MUI builds (multiple languages on the same machine)
  //and Windows Vista. For more details on this, please see Larry Osterman's WebLog at https://docs.microsoft.com/en-us/archive/blogs/larryosterman/a-peek-behind-the-beep
  if (m_nAppFriendlyNameID != 0)
  {
    //Form the indirect shell string
    ATL::CAtlString sIndirectShellString{FormIndirectShellString(sResourceModule, m_nAppFriendlyNameID)};
    lResult = appKey.SetStringValue(_T("DispFileName"), sIndirectShellString, REG_SZ);
    if (lResult != ERROR_SUCCESS)
      return ATL::AtlHresultFromWin32(lResult);
  }

  //Now write out all the individual sounds
#pragma warning(suppress: 26477)
  ATLASSERT(m_nSounds); //We need at least one sound!
  for (int i{0}; i<m_nSounds; i++)
  {
    const HRESULT hr{RegisterSound(appKey, eventLabelsKey, i, bFlush, sAppPath, sResourceModule)};
    if (FAILED(hr))
      return hr;
  }

  //Flush the registry if required
  if (bFlush)
  {
    lResult = eventLabelsKey.Flush();
    if (lResult != ERROR_SUCCESS)
      return ATL::AtlHresultFromWin32(lResult);
    lResult = appKey.Flush();
    if (lResult != ERROR_SUCCESS)
      return ATL::AtlHresultFromWin32(lResult);
  }

  return S_OK;
}

HRESULT CAppSoundsManager::UnRegister(_In_ bool bFlush)
{
  //Also try to open the HKCU\AppEvents\EventLabels key
  ATL::CRegKey eventLabelsKey;
  LSTATUS lResult{eventLabelsKey.Open(HKEY_CURRENT_USER, _T("AppEvents\\EventLabels"), KEY_WRITE)};
  if (lResult != ERROR_SUCCESS)
    return ATL::AtlHresultFromWin32(lResult);

  //Now delete out all the individual sounds from "HKCU\AppEvents\EventLabels"
  for (int i{0}; i<m_nSounds; i++)
  {
    //Delete the key under "HKCU\AppEvents\Schemes\Apps\[Alias]"
#pragma warning(suppress: 26481)
    ATL::CAtlString sAlias{FormAlias(m_pSounds[i])};
    lResult = eventLabelsKey.RecurseDeleteKey(sAlias);
    if (lResult != ERROR_SUCCESS)
      return ATL::AtlHresultFromWin32(lResult);
  }

  //Flush the registry if required
  if (bFlush)
  {
    lResult = eventLabelsKey.Flush();
    if (lResult != ERROR_SUCCESS)
      return ATL::AtlHresultFromWin32(lResult);
  }

  //Finally delete the HKCU\AppEvents\Schemes\Apps\<FILENAME> key
  ATL::CRegKey currentUserKey;
  lResult = currentUserKey.Open(HKEY_CURRENT_USER, nullptr, KEY_WRITE);
  if (lResult != ERROR_SUCCESS)
    return ATL::AtlHresultFromWin32(lResult);
  lResult = currentUserKey.RecurseDeleteKey(GetAppKeyName());
  if (lResult != ERROR_SUCCESS)
    return ATL::AtlHresultFromWin32(lResult);

  return S_OK;
}

HRESULT CAppSoundsManager::PlaySound(_In_ UINT nSoundID, _In_ DWORD dwFlags)
{
  if (nSoundID == 0) //A Sound ID of 0 has a special meaning of stopping any sound which is currently playing
  {
    if (!::PlaySound(nullptr, nullptr, 0))
      return ATL::AtlHresultFromLastError();
  }
  else
  {
    //Lookup the current value for the specified sound in the registry
    ATL::CAtlString sFilename;
    const HRESULT hr{GetCurrentSoundFromRegistry(nSoundID, sFilename)};
    if (SUCCEEDED(hr))
    {
      if (!::PlaySound(sFilename, nullptr, SND_FILENAME | dwFlags))
        return ATL::AtlHresultFromLastError();
    }
    else
    {
      //Ok, nothing in the registry, see if we should play the sound from a resource
      const CAppSound* pFound{FindAppSound(nSoundID)};
      if (pFound != nullptr)
      {
        if (pFound->m_bPlayFromResource)
        {
#pragma warning(suppress: 26477)
          ATLASSERT(m_hResource);
          if (!::PlaySound(MAKEINTRESOURCE(nSoundID), m_hResource, SND_RESOURCE | dwFlags))
            return ATL::AtlHresultFromLastError();
        }
        else
          return ATL::AtlHresultFromWin32(ERROR_RESOURCE_DISABLED);
      }
      else
        return ATL::AtlHresultFromWin32(ERROR_FILE_NOT_FOUND);
    }
  }

  return S_OK;
}
