/*
Module : AppSoundsManager.h
Purpose: Defines the interface for CAppSoundsManager, a class to encapsulate "Applications Sounds".
         The class is based in part upon the "CSoundMgr" class developed by Paul DiLascia for the 
         May 2006 edition of the MSDN Magazine
Created: PJN / 22-09-2006

Copyright (c) 2006 - 2020 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


/////////////////////////////// Macros / Defines //////////////////////////////

#pragma once

#ifndef __APPSOUNDSMANAGER_H__
#define __APPSOUNDSMANAGER_H__

#ifndef SOUNDSMANAGER_EXT_CLASS
#define SOUNDSMANAGER_EXT_CLASS
#endif //#ifndef SOUNDSMANAGER_EXT_CLASS

#ifndef __ATLSTR_H__
#pragma message("To avoid this message, please put atlstr.h in your pre compiled header (normally stdafx.h)")
#include <atlstr.h>
#endif //#ifndef __ATLSTR_H__


/////////////////////////////// Classes ///////////////////////////////////////

//The class which represents one specific application sound
class SOUNDSMANAGER_EXT_CLASS CAppSound
{
public:
  UINT    m_nID;               //The programmatic ID of the sound. This represents the id when registered into the registry, as well as the resource 
                               //id of the WAVE file to play and also the string resource id representing its friendly name
  LPCTSTR m_pszPath;           //The path for the WAVE file to use when registering. It can include the replaceable parameter %APPPATH% similar to 
                               //the %MODULE% value ATL rgs scripts. For example if you wanted to register a sound to play from the directory where 
                               //you exe is located you would use something like "%APPPATH%\MySoundFile.wav"
  bool    m_bPlayFromResource; //Should we play this sound from a resource if no value is found for it in the registry
};

//The main class which manages an Applications set of sounds
class SOUNDSMANAGER_EXT_CLASS CAppSoundsManager
{
public:
//Constructors / Destructors
  CAppSoundsManager() noexcept;
  CAppSoundsManager(const CAppSoundsManager&) = delete;
  CAppSoundsManager(CAppSoundsManager&&) = delete;
  virtual ~CAppSoundsManager() = default;

//Methods
  CAppSoundsManager* operator=(const CAppSoundsManager&) = delete;
  CAppSoundsManager* operator=(CAppSoundsManager&&) = delete;
  virtual void SetupSoundsTable() = 0;                                                                                   //Implemented by the sound map macros
  void SetupSounds(_In_ const CAppSound* pSounds, _In_ int nSounds) noexcept;                                            //Sets up the sound map, called as part of END_APP_SOUNDS
  virtual HRESULT Initialize(_In_z_ LPCTSTR pszAppFriendlyName, _In_ UINT nAppFriendlyNameID, _In_ HINSTANCE hResource); //Setups the class with important settings
  HRESULT PlaySound(_In_ UINT nSoundID, _In_ DWORD dwFlags = 0);                                                         //Plays the specified app sound
  const CAppSound* FindAppSound(_In_ UINT nSoundID) noexcept;                                                            //Find the specified sound in our app sounds array
  HRESULT GetCurrentSoundFromRegistry(_In_ UINT nSoundID, _Inout_ ATL::CAtlString& sSoundFileName);                      //Find the current value associated with an app sound from the registry
  HRESULT IsRegistered();
  HRESULT Register(_In_ bool bFlush = false);
  HRESULT UnRegister(_In_ bool bFlush = false);
  void SetEnable(_In_ bool bEnable) noexcept { m_bEnabled = bEnable; };
  _NODISCARD bool GetEnable() const noexcept { return m_bEnabled; };
  void SetAppFriendlyName(_In_z_ LPCTSTR pszAppFriendlyName) { m_sAppFriendlyName = pszAppFriendlyName; };
  void SetAppFriendlyName(_In_ UINT nResourceID) noexcept { m_nAppFriendlyNameID = nResourceID; };
  _NODISCARD ATL::CAtlString GetAppFriendlyName() const { return m_sAppFriendlyName; };
  void SetAppFileName(_In_z_ LPCTSTR pszFilename) { m_sFilename = pszFilename; };
  _NODISCARD ATL::CAtlString GetAppFileName() const { return m_sFilename; };

protected:
//Methods
  virtual ATL::CAtlString GetAppKeyName();
  virtual ATL::CAtlString GetSoundAliasCurrentValueKeyName(_In_ UINT nSoundID);
  virtual HRESULT RegisterSound(_In_ HKEY hAppKey, _In_ HKEY hEventLabelsKey, _In_ int nSoundIndex, _In_ bool bFlush, _In_z_ LPCTSTR pszAppPath, _In_z_ LPCTSTR pszResourceModule);
  virtual ATL::CAtlString FormAlias(_In_ const CAppSound& sound);
  virtual ATL::CAtlString FormIndirectShellString(_In_z_ LPCTSTR pszResourceModule, _In_ UINT nID);

//Member variables
  int m_nSounds;                      //The size of the m_pAppSounds array
  const CAppSound* m_pSounds;         //All the sounds for this application
  bool m_bEnabled;                    //Are we currently enabled to play sounds
  HINSTANCE m_hResource;              //The resource handle to use for playing sounds from resources (By default will be AfxGetResourceHandle())
  ATL::CAtlString m_sAppFriendlyName; //The friendly application name (By default will be AfxGetAppName())
  UINT m_nAppFriendlyNameID;          //The resource ID we will use for the friendly application name (by default will be AFX_IDS_APP_TITLE)
  ATL::CAtlString m_sFilename;        //The value we use for the FILENAME when we do the registration to HKCU\AppEvents\Schemes\Apps\<FILENAME>
};


//////////////////////////// Macros ///////////////////////////////////////////

#define BEGIN_APP_SOUNDS() \
  static const CAppSound* GetSoundsMap(int* pCount) noexcept \
  { \
    static CAppSound sm_Sounds[] = /*NOLINT(modernize-avoid-c-arrays)*/ \
    {

#define APP_SOUND_FROM_RESOURCE(SoundID) \
      APP_SOUND_EX(SoundID, nullptr, 1)

#define APP_SOUND_FROM_FILE(SoundID, RegisteredPath) \
      APP_SOUND_EX(SoundID, RegisteredPath, 0)

#define APP_SOUND_DEFAULT_NONE(SoundID) \
      APP_SOUND_EX(SoundID, nullptr, 0)

#define APP_SOUND_EX(SoundID, RegisteredPath, PlayFromResource) \
      { SoundID, RegisteredPath, PlayFromResource },

#define END_APP_SOUNDS() \
    }; \
    \
    if (pCount != nullptr) \
      *pCount = sizeof(sm_Sounds)/sizeof(CAppSound); \
    return sm_Sounds; \
  }; \
  \
  void SetupSoundsTable() noexcept override \
  { \
    int nCount = 0; \
    const CAppSound* pSounds = GetSoundsMap(&nCount); \
    SetupSounds(pSounds, nCount); \
  };

#endif //#ifndef __APPSOUNDSMANAGER_H__
